from threading import Thread, Event
import requests
import time
import logging
logger = logging.getLogger(__name__)


class BfgStatsPoller(Thread):
    def __init__(self, port):
        super(BfgStatsPoller, self).__init__()
        self.stopped = Event()
        self.buffer = []
        self.port = port

    def run(self):
        last_ts = int(time.time() - 1)
        while not self.stopped.is_set():
            curr_ts = int(time.time())
            if curr_ts > last_ts:
                last_ts = curr_ts
                try:
                    bfg_stat = requests.get(
                        "http://localhost:{port}/stats".format(port=self.port), timeout=0.9
                    ).json()
                    data = {
                        'ts': last_ts - 1,
                        'metrics': {
                            'instances': bfg_stat.get('instances'),
                            'reqps': bfg_stat.get('reqps')
                        }
                    }
                except (requests.ConnectionError, requests.HTTPError, requests.exceptions.Timeout):
                    logger.debug("Bfg http stat server interface is unavailable", exc_info=True)
                    data = {
                        'ts': last_ts - 1,
                        'metrics': {
                            'instances': 0,
                            'reqps': 0
                        }
                    }
                self.buffer.append(data)
            else:
                time.sleep(0.2)

    def stop(self):
        self.stopped.set()

    def get_data(self):
        result, self.buffer = self.buffer, []
        return result


class BfgStatsReader(object):
    def __init__(self, port):
        self.closed = False
        self.port = port
        self.poller = BfgStatsPoller(port)
        self.started = False

    def __next__(self):
        if self.closed:
            raise StopIteration()
        elif not self.started:
            self.poller.start()
            self.started = True
        return self.poller.get_data()

    def close(self):
        self.closed = True
        self.poller.stop()
        self.poller.join()

    def __iter__(self):
        return self



import ctypes
import logging
import os
import shutil
import time
from configparser import RawConfigParser, MissingSectionHeaderError
from multiprocessing import Manager, Event, Value, Process

import glob
import stat
import yaml
from pkg_resources import resource_filename

from yandextank.common.interfaces import TankInfo
from yandextank.common.util import Cleanup, Finish, Status
from yandextank.common.util import read_resource, TankapiLogFilter
from yandextank.config_converter.converter import convert_ini, convert_single_option
from yandextank.core import TankCore
from yandextank.core.tankcore import LockError, Lock
from yandextank.validator.validator import ValidationError

logger = logging.getLogger()


class TankWorker(Process):
    SECTION = 'core'
    FINISH_FILENAME = 'finish_status.yaml'
    DEFAULT_CONFIG = 'load.yaml'

    def __init__(self, configs, cli_options=None, cfg_patches=None, cli_args=None, no_local=False,
                 log_handlers=None, wait_lock=False, files=None, ammo_file=None,
                 debug=False, run_shooting_event=None, storage=None):
        super().__init__()
        self.interrupted = Event()
        manager = Manager()
        self.info = TankInfo(manager.dict())
        self.config_list = self._combine_configs(configs, cli_options, cfg_patches, cli_args, no_local)
        self.core = TankCore(self.config_list, self.interrupted, self.info, storage=storage)

        is_locked = Lock.is_locked(self.core.lock_dir)
        if is_locked and not self.core.config.get_option(self.SECTION, 'ignore_lock'):
            raise LockError(is_locked)

        self.wait_lock = wait_lock
        self.log_handlers = log_handlers if log_handlers is not None else []
        self.files = [] if files is None else files
        self.ammo_file = ammo_file
        self.config_paths = configs
        self.folder = self.core.artifacts_dir
        self.init_logging(debug or self.core.get_option(self.core.SECTION, 'debug'))

        self._status = Value(ctypes.c_char_p, Status.TEST_INITIATED)
        self._test_id = Value(ctypes.c_char_p, self.core.test_id.encode('utf8'))
        self._retcode = Value(ctypes.c_int, 0)
        self._msgs = manager.list()
        self._run_shooting_event = run_shooting_event or self._dummy_event()

    @staticmethod
    def _dummy_event():
        event = Event()
        event.set()
        return event

    def run(self):
        with Cleanup(self) as add_cleanup:
            lock = self.get_lock()
            add_cleanup('release lock', lock.release)
            self.status = Status.TEST_PREPARING
            logger.info('Created a folder for the test. %s' % self.folder)
            self.core.plugins_configure()
            add_cleanup('plugins cleanup', self.core.plugins_cleanup)
            self.core.plugins_prepare_test()
            with Finish(self):
                if not self._run_shooting_event.is_set():
                    self.status = Status.TEST_WAITING_FOR_A_COMMAND_TO_RUN
                    self._wait_for_a_command_to_start_shooting()
                self.status = Status.TEST_RUNNING
                self.core.plugins_start_test()
                self.retcode = self.core.wait_for_finish()
                self._msgs.extend(self.core.errors)
            self.status = Status.TEST_POST_PROCESS
            self.retcode = self.core.plugins_post_process(self.retcode)

    def _wait_for_a_command_to_start_shooting(self):
        pool_timeout = 0.01
        while True:
            if self.core.interrupted.is_set():
                raise RuntimeError('Test stopped before shooting started.')
            if self._run_shooting_event.wait(pool_timeout):
                return

    @staticmethod
    def _combine_configs(run_cfgs, cli_options=None, cfg_patches=None, cli_args=None, no_local=False):
        if cli_options is None:
            cli_options = []
        if cfg_patches is None:
            cfg_patches = []
        if cli_args is None:
            cli_args = []
        run_cfgs = run_cfgs if len(run_cfgs) > 0 else [TankWorker.DEFAULT_CONFIG]

        if no_local:
            configs = [load_cfg(cfg) for cfg in run_cfgs] + \
                parse_options(cli_options) + \
                parse_and_check_patches(cfg_patches) + \
                cli_args
        else:
            configs = [load_core_base_cfg()] + \
                load_local_base_cfgs() + \
                [load_cfg(cfg) for cfg in run_cfgs] + \
                parse_options(cli_options) + \
                parse_and_check_patches(cfg_patches) + \
                cli_args
        return configs

    def stop(self):
        self.interrupted.set()
        logger.warning('Interrupting')

    def get_status(self):
        return {'status_code': self.status.decode('utf8'),
                'left_time': None,
                'exit_code': self.retcode,
                'lunapark_id': self.get_info('uploader', 'job_no'),
                'tank_msg': self.msg,
                'test_id': self.test_id,
                'lunapark_url': self.get_info('uploader', 'web_link')
                }

    def save_finish_status(self):
        with open(os.path.join(self.folder, self.FINISH_FILENAME), 'w') as f:
            yaml.safe_dump(self.get_status(), f, encoding='utf-8', allow_unicode=True)

    def get_info(self, section_name, key_name):
        return self.info.get_value([section_name, key_name])

    def init_logging(self, debug=False):

        filename = os.path.join(self.core.artifacts_dir, 'tank.log')
        open(filename, 'a').close()
        current_file_mode = os.stat(filename).st_mode
        os.chmod(filename, current_file_mode | stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH)

        logger.handlers = []
        logger.setLevel(logging.DEBUG if debug else logging.INFO)

        file_handler = logging.FileHandler(filename)
        file_handler.setLevel(logging.DEBUG)
        file_handler.setFormatter(logging.Formatter(
            "%(asctime)s [%(levelname)s] %(name)s %(filename)s:%(lineno)d\t%(message)s"))
        file_handler.addFilter(TankapiLogFilter())
        logger.addHandler(file_handler)
        logger.info("Log file created")

        for handler in self.log_handlers:
            logger.addHandler(handler)
            logger.info("Logging handler {} added".format(handler))

    def get_lock(self):
        while not self.interrupted.is_set():
            try:
                lock = Lock(self.test_id, self.folder).acquire(self.core.lock_dir,
                                                               self.core.config.get_option(self.SECTION, 'ignore_lock'))
                break
            except LockError as e:
                self.add_msgs(str(e))
                if not self.wait_lock:
                    raise RuntimeError("Lock file present, cannot continue")
                logger.warning(
                    "Couldn't get lock. Will retry in 5 seconds...")
                time.sleep(5)
        else:
            raise KeyboardInterrupt
        return lock

    @property
    def msg(self):
        return '\n'.join(self._msgs)

    def add_msgs(self, *msgs):
        self._msgs.extend(msgs)

    @property
    def test_id(self):
        with self._test_id.get_lock():
            return self._test_id.value.decode('utf8')

    @property
    def status(self):
        with self._status.get_lock():
            return self._status.value

    @status.setter
    def status(self, val):
        with self._status.get_lock():
            self._status.value = val

    @property
    def retcode(self):
        with self._retcode.get_lock():
            return self._retcode.value

    @retcode.setter
    def retcode(self, val):
        with self._retcode.get_lock():
            self._retcode.value = val

    def collect_files(self):
        for cfg in self.config_paths:
            shutil.move(cfg, self.folder)
        for f in self.files:
            shutil.move(f, self.folder)
        if self.ammo_file:
            shutil.move(self.ammo_file, self.folder)

    def go_to_test_folder(self):
        os.chdir(self.folder)


def load_cfg(cfg_filename):
    """

    :type cfg_filename: str
    """
    if is_ini(cfg_filename):
        return convert_ini(cfg_filename)
    else:
        cfg_yaml = yaml.load(read_resource(cfg_filename), Loader=yaml.FullLoader)
        if not isinstance(cfg_yaml, dict):
            raise ValidationError('Wrong config format, should be a yaml')
        return cfg_yaml


def load_core_base_cfg():
    return load_cfg(resource_filename(__name__, 'config/00-base.yaml'))


def load_local_base_cfgs():
    return cfg_folder_loader('/etc/yandex-tank')


def parse_options(options):
    """
    :type options: list of str
    :rtype: list of dict
    """
    if options is None:
        return []
    else:
        return [
            convert_single_option(key.strip(), value.strip())
            for key, value
            in [option.split('=', 1) for option in options]
        ]


def parse_and_check_patches(patches):
    parsed = [yaml.load(p, Loader=yaml.FullLoader) for p in patches]
    for patch in parsed:
        if not isinstance(patch, dict):
            raise ValidationError('Config patch "{}" should be a dict'.format(patch))
    return parsed


def cfg_folder_loader(path):
    """
    :type path: str
    """
    CFG_WILDCARD = '*.yaml'
    return [load_cfg(filename) for filename in sorted(glob.glob(os.path.join(path, CFG_WILDCARD)))]


def is_ini(cfg_file):
    if cfg_file.endswith('.yaml') or cfg_file.endswith('.json'):
        return False
    try:
        RawConfigParser().read(cfg_file)
        return True
    except MissingSectionHeaderError:
        return False



# TODO: consider removing this module from library. It is very specific.
import logging
import weakref
import usb1
import libusb1

logger = logging.getLogger(__name__)


class UsbHandler(object):
    """ Usb handler"""

    ANDROID_USB_SETTINGS = (0xFF, 0x42, 0x01)  # (class, subclass, proto), adb.h
    APPLE_VENDOR_ID = 0x05ac
    IPHONE_PRODUCT_ID = 0x12a8
    IPAD_PRODUCT_ID = 0x12ab

    def __init__(self, source=None, device_type='android', timeout=10000):
        super(UsbHandler, self).__init__()
        self.timeout = timeout
        self.source = source
        self.source_type = 'usb' if self.source.startswith('usb:') else 'serial'
        self.device_type = device_type
        self.interface_number = None
        self.__max_read_packet_len = 0
        self.device = None
        self.settings = None
        self.handle = None

    def initialize(self):
        self.device, self.settings = self.get_device()
        self.__get_endpoints()
        self.open()

    @staticmethod
    def _get_devices_by_usb_settings(related_usb_settings):
        ctx = usb1.USBContext()
        return [
            (device, settings)
            for device in ctx.getDeviceList(skip_on_error=True) for settings in device.iterSettings()
            if (settings.getClass(), settings.getSubClass(), settings.getProtocol()) == related_usb_settings
        ]

    @staticmethod
    def _get_devices_by_product_and_vendor_id(vendor_id, product_id):
        ctx = usb1.USBContext()
        return [
            (device, None)
            for device in ctx.getDeviceList(skip_on_error=True)
            if device.getVendorID() == vendor_id and device.getProductID() == product_id
        ]

    def get_devices(self):
        return {
            'ipdas': self._get_devices_by_product_and_vendor_id(self.APPLE_VENDOR_ID, self.IPAD_PRODUCT_ID),
            'iphones': self._get_devices_by_product_and_vendor_id(self.APPLE_VENDOR_ID, self.IPHONE_PRODUCT_ID),
            'androids': self._get_devices_by_usb_settings(self.ANDROID_USB_SETTINGS)
        }

    def get_device(self):
        devices_list = self._get_devices_by_usb_settings(self.ANDROID_USB_SETTINGS)
        logger.debug('Related devices list: %s', devices_list)
        matched_devices = None
        if devices_list:
            if self.source_type == 'usb':
                matched_devices = [
                    (device, settings) for device, settings in devices_list
                    if [device.getBusNumber()] + device.getPortNumberList() == self.source
                ]
            elif self.source_type == 'serial':
                matched_devices = [
                    (device, settings) for device, settings in devices_list
                    if device.getSerialNumber() == self.source
                ]
            else:
                matched_devices = []
        else:
            raise ValueError('No devices found. Is it plugged in and in appropriate state?')
        logger.debug('Matched devices list: %s', matched_devices)
        if not matched_devices:
            raise ValueError('No suitable devices found for source: %s', self.source)
        if len(matched_devices) > 1:
            raise ValueError(
                'There are more than 1 devices found for this source: %s. Devices: %s',
                self.source, matched_devices
            )
        return matched_devices[0]

    def __get_endpoints(self):
        for endpoint in self.settings.iterEndpoints():
            address = endpoint.getAddress()
            if address & libusb1.USB_ENDPOINT_DIR_MASK:
                self.__read_endpoint = address
                self.__max_read_packet_len = endpoint.getMaxPacketSize()
            else:
                self.__write_endpoint = address
        if not self.__read_endpoint or not self.__write_endpoint:
            raise RuntimeError('USB endpoints not found')

    def open(self):
        handle = self.device.open()
        interface = self.settings.getNumber()
        if handle.kernelDriverActive(interface):
            handle.detachKernelDriver(interface)
        handle.claimInterface(interface)
        self.handle = handle
        self.interface_number = interface
        logger.debug('Opened usb handler: %s. Iface: %s', self.handle, self.interface_number)
        weakref.ref(self, self.close)  # When this object is deleted, make sure it's closed.

    def close(self):
        logger.info('Closing handle...')
        try:
            self.handle.releaseInterface(self.interface_number)
            self.handle.close()
        except usb1.USBError:
            logger.warning('USBError while closing handle %s:', exc_info=True)
        finally:
            self.handle = None

    def flush(self):
        while True:
            try:
                self.read(self.__max_read_packet_len)
            except ReadFailedError as e:
                if e.usb_error.value == libusb1.LIBUSB_ERROR_TIMEOUT:
                    break
                raise

    def write(self, data):
        if self.handle is None:
            raise WriteFailedError(
                'This handle has been closed, probably due to another being opened.',
                None)
        try:
            self.handle.bulkWrite(self.__write_endpoint, data, timeout=self.timeout)
        except usb1.USBError:
            logger.warning('Usb write failed, data: %s', data, exc_info=True)
            raise

    def read(self, length):
        if self.handle is None:
            raise ReadFailedError(
                'This handle has been closed, probably due to another being opened.',
                None)
        try:
            chunk = self.handle.bulkRead(self.__read_endpoint, length, timeout=self.timeout)
        except usb1.USBError:
            logger.warning('Usb read failed', exc_info=True)
            raise
        else:
            return bytearray(chunk)


class CommonUsbError(Exception):
    """Base class for handlers communication errors."""


class LibusbWrappingError(CommonUsbError):
    """Wraps libusb1 errors while keeping its original usefulness.

    Attributes:
      usb_error: Instance of libusb1.USBError
    """

    def __init__(self, msg, usb_error):
        super(LibusbWrappingError, self).__init__(msg)
        self.usb_error = usb_error

    def __str__(self):
        return '%s: %s' % (
            super(LibusbWrappingError, self).__str__(), str(self.usb_error))


class WriteFailedError(LibusbWrappingError):
    """Raised when the device doesn't accept our command."""


class ReadFailedError(LibusbWrappingError):
    """Raised when the device doesn't respond to our commands."""



import re

import time
from queue import Empty

from multiprocessing import Queue, Event, Process

import logging

import signal
from yandextank.common.interfaces import MonitoringPlugin
from infra.yasm.yasmapi import RtGolovanRequest
from threading import Thread

logger = logging.getLogger(__name__)


DEFAULT_SIGNALS = {
    'portoinst-cpu_usage_cores_tmmv': 'cpu_usage',
    'portoinst-cpu_guarantee_cores_tmmv': 'cpu_guarantee',
    'portoinst-cpu_limit_cores_tmmv': 'cpu_limit',
    'portoinst-cpu_wait_cores_tmmv': 'cpu_wait',
    'portoinst-memory_usage_gb_tmmv': 'memory_usage',
    'portoinst-memory_limit_gb_tmmv': 'memory_limit',
    'portoinst-io_read_fs_bytes_tmmv': 'io_read',
    'portoinst-io_write_fs_bytes_tmmv': 'io_write',
    'portoinst-io_limit_bytes_tmmv': 'io_limit',
    'conv(unistat-auto_disk_rootfs_usage_bytes_axxx, Gi)': 'rootfs_usage',
    'conv(unistat-auto_disk_rootfs_total_bytes_axxx, Gi)': 'rootfs_total',
    'portoinst-net_mb_summ': 'net_mb',
    'portoinst-net_guarantee_mb_summ': 'net_guarantee',
    'portoinst-net_limit_mb_summ': 'net_limit'
}


def signals_stream(panel):
    '''
    :type yasmapi_cfg: Panel
    :return: Panel, float, dict
    '''
    for point in RtGolovanRequest(panel.as_dict):
        yield point.ts, point.values[panel.host][panel.tags]


def map_metric_name(name):
    NAME_MAP = {r'conv\(.+\)': r'(?<=conv\().+?(?=,)'}
    for pattern, mask in NAME_MAP.items():
        if re.match(pattern, name):
            name = re.findall(mask, name)[0]
            break
    return 'custom:{}'.format(name)


def convert_value(name, value):
    return value


def monitoring_data(ts, data, metric_map, comment=''):
    return {
        "timestamp": ts,
        "data": {
            host: {
                "comment": comment,
                "metrics": {map_metric_name(metric_map[name]): convert_value(name, value) for name, value in host_data.items()}
            }
            for host, host_data in data.items()}}


class YasmCfg(object):
    def __init__(self, panels):
        self.panels = [Panel(alias, **attrs) for alias, attrs in panels.items()]
        self._as_dict = None

    @property
    def as_dict(self):
        if self._as_dict is None:
            yasmapi_cfg = {}
            for panel in self.panels:
                yasmapi_cfg.setdefault(panel.host, {})[panel.tags] = panel.signals
            logger.info('yasmapi cfg: {}'.format(yasmapi_cfg))
            self._as_dict = yasmapi_cfg
        return self._as_dict


class Panel(object):
    def __init__(self, alias, host, tags, signals=None, default_signals=True):
        self.queue = Queue()
        self.alias = alias
        self.extend_signals = self._check_signals_(signals)
        self.signals = [signal.get('metric', '') if isinstance(signal, dict) else signal for signal in self.extend_signals] + (list(DEFAULT_SIGNALS.keys()) if default_signals else [])

        self.host = host
        self.tags = tags.strip(';')

        if len(self.signals) == 0:
            logger.warning('No signals specified for {} panel'.format(self.alias))
        self.as_dict = {self.host: {self.tags: self.signals}}
        self.metric_map = self._metrics_map_()

    def _check_signals_(self, signals):
        checked = []
        if isinstance(signals, list):
            for element in signals:
                if isinstance(element, str):
                    checked.append(element)
                elif isinstance(element, dict):
                    if element.get('metric'):
                        checked.append(element)
                    else:
                        logging.warning('Wrong signal: %s', element)
                else:
                    logging.warning('Wrong signal: %s', element)
        return checked

    def _metrics_map_(self):
        metrics_map = DEFAULT_SIGNALS
        assign_alias = set(metrics_map.values())
        for metric in self.extend_signals:
            key = metric.get('metric', '') if isinstance(metric, dict) else metric
            mark = metric.get('mark', '') if isinstance(metric, dict) else None
            value = '{}_signal'.format(mark) if mark and mark not in assign_alias else key
            if key:
                metrics_map.update({key: value})
        return metrics_map


class YasmMPReceiver(object):
    def __init__(self, yasm_cfg, yasmapi_timeout):
        """

        :type data_queue: Queue
        :type yasm_cfg: YasmCfg
        """
        self.panels = yasm_cfg.panels
        self.data_queue = Queue()
        self.timeout = yasmapi_timeout
        self._data_buffer = []
        self.start_event = Event()
        self.stop = Event()
        self.interrupted = Event()
        self.end_time = None
        self.ps_pool = {panel.alias: Process(target=self.single_receiver,
                                             args=(panel,))
                        for panel in self.panels}
        self.consumers = {panel.alias: Thread(target=self.single_consumer, args=(panel, self.ps_pool[panel.alias]))
                          for panel in self.panels}

    def get_buffered_data(self):
        data, self._data_buffer = self._data_buffer, []
        return data

    def start_collecting(self):
        [p.start() for p in self.ps_pool.values()]

    def start_transmitting(self):
        self.start_event.set()
        [consumer.start() for consumer in self.consumers.values()]

    def single_receiver(self, panel):
        # ignore SIGINT (process is controlled by .stop_event)
        """

        :type panel: Panel
        """
        signal.signal(signal.SIGINT, signal.SIG_IGN)

        stream = signals_stream(panel)
        try:
            while not self.stop.is_set():
                ts, data = next(stream)
                if self.start_event.is_set():
                    panel.queue.put((ts, {panel.alias: data}))
            else:
                logger.info('Stopped collecting metrics for panel {}'.format(panel.alias))
        except KeyboardInterrupt:
            pass
        finally:
            logger.info('Closing panel {} receiver thread'.format(panel.alias))

    def single_consumer(self, panel, ps):
        """

        :type ps: Process
        :type panel: Panel
        """
        while not self.interrupted.is_set():
            try:
                ts, data = panel.queue.get(timeout=self.timeout)
                self._data_buffer.append(monitoring_data(ts, data, panel.metric_map))
                if self.end_time is not None:
                    if ts > self.end_time:
                        logger.info('{} all yasm metrics received'.format(panel.alias))
                        self.stop.set()
                        break
                    else:
                        logger.info('Waiting for yasm metrics for panel {}'.format(panel.alias))
                else:
                    logger.debug('New ts: {}'.format(ts))
            except Empty:
                logger.warning(
                    'Not receiving any data for panel {} from YASM.'
                    'Probably your hosts/tags specification is not correct'.format(panel.alias))
                if ps.is_alive():
                    ps.terminate()
                break
        else:
            logger.info('Interrupting {}'.format(panel.alias))
            self.stop.set()
        logger.debug("Stopping panel {} controller")

    def stop_now(self):
        self.end_time = time.time()
        logger.info('Stopping yasm plugin')
        try:
            for panel in self.panels:
                if self.consumers[panel.alias].is_alive():
                    self.consumers[panel.alias].join()
                if self.ps_pool[panel.alias].is_alive():
                    self.ps_pool[panel.alias].terminate()
        except KeyboardInterrupt:
            self.interrupted.set()
            for panel in self.panels:
                if self.consumers[panel.alias].is_alive():
                    self.consumers[panel.alias].join()
                if self.ps_pool[panel.alias].is_alive():
                    self.ps_pool[panel.alias].terminate()


class Plugin(MonitoringPlugin):
    def __init__(self, core, cfg, name):
        super(Plugin, self).__init__(core, cfg, name)
        self.yasm_receiver = YasmMPReceiver(YasmCfg(self.get_option('panels')),
                                            self.get_option('timeout'))

    def send_collected_data(self, data):
        """sends pending data set to listeners"""
        self.core.publish_monitoring_data(data)
        # for listener in self.listeners:
        #     listener.monitoring_data(data)

    def is_test_finished(self):
        data = self.yasm_receiver.get_buffered_data()
        if len(data) > 0:
            self.send_collected_data(data)
        return -1

    def prepare_test(self):
        self.yasm_receiver.start_collecting()

    def start_test(self):
        self.yasm_receiver.start_transmitting()

    def end_test(self, retcode):
        self.yasm_receiver.stop_now()
        self.send_rest()
        return retcode

    def send_rest(self):
        data = self.yasm_receiver.get_buffered_data()
        if len(data) > 0:
            self.send_collected_data(data)



from yandextank.common.interfaces import MonitoringPlugin
from .solomonreceiver import SolomonReceiver
from logging import getLogger
from re import match


logger = getLogger(__name__)


def set_timeout(timestring):
    if match(r"^\d+s$", timestring):
        return int(timestring[:-1])
    elif match(r"^\d+m$", timestring):
        return 60 * int(timestring[:-1])
    elif match(r"^\d+h$", timestring):
        return 3600 * int(timestring[:-1])
    else:
        return 30


class Plugin(MonitoringPlugin):

    def __init__(self, core, cfg, name):
        super(Plugin, self).__init__(core, cfg, name)
        try:
            self.timeout = int(self.get_option('timeout'))
        except ValueError:
            self.timeout = set_timeout(self.get_option('timeout'))

    def send_collected_data(self, data):
        self.core.publish_monitoring_data(data)

    def is_test_finished(self):   # noqa: PLE0202
        data = self.solomon.get_buffered_data()
        if len(data) > 0:
            for metric in data:
                self.send_collected_data([metric])
        return -1

    def prepare_test(self):
        try:
            with open(self.get_option('token_file'), 'r') as tfile:
                token = "{}".format(tfile.read()).strip("\n")
        except (OSError, IOError):
            error = "Solomon plugin: Authorization token is not set! File {} is not found or can't to read.".format(self.get_option('token_file'))
            logger.warning(error)
            token = None
        if token:
            self.solomon = SolomonReceiver(self.get_option('api_host'), self.get_option('panels'), self.timeout, token)
            self.solomon.prepare()
        elif self.get_option('enforce_check_token'):
            raise RuntimeError(error)
        else:
            self.start_test = lambda *a, **kw: None
            self.is_test_finished = self.stub
            self.end_test = self.stub

    def start_test(self):   # noqa: PLE0202
        self.solomon.run()

    def end_test(self, retcode):   # noqa: PLE0202
        self.solomon.stop_now()
        self.send_rest()
        return retcode

    def send_rest(self):
        data = self.solomon.get_buffered_data()
        if len(data) > 0:
            for metric in data:
                self.send_collected_data([metric])

    def stub(self, *a, **kwa):
        return -1



''' Plugin that can be used to store options for adjacent CI or scripts in tank config '''
import logging

from ...common.interfaces import AbstractPlugin


logger = logging.getLogger(__name__)


class Plugin(AbstractPlugin):
    """   MetaConf plugin to store and validate options     """

    def __init__(self, core, cfg, name):
        AbstractPlugin.__init__(self, core, cfg, name)
        self.config_content = self.core.config.validated["metaconf"]

    @staticmethod
    def get_key():
        return __file__

    def configure(self):
        logger.debug("Meta config used in test:\n%s", self.config_content)



import logging

from urllib.parse import urljoin

import re
import pandas
import requests
from yandextank.contrib.netort.netort.data_manager import DataSession, thread_safe_property
import threading as th
from requests import ConnectionError

from yandextank.plugins.Phantom.reader import string_to_df_microsec
from yandextank.common.interfaces import AbstractPlugin,\
    MonitoringDataListener

logger = logging.getLogger(__name__)  # pylint: disable=C0103


class Plugin(AbstractPlugin, MonitoringDataListener):
    SECTION = 'neuploader'
    importance_high = {
        'interval_real',
        'proto_code',
        'net_code'
    }
    OVERALL = '__overall__'
    LUNA_LINK = 'https://luna.yandex-team.ru/tests/'
    PLANNED_RPS_METRICS_NAME = 'planned_rps'
    ACTUAL_RPS_METRICS_NAME = 'actual_rps'

    def __init__(self, core, cfg, name):
        super(Plugin, self).__init__(core, cfg, name)
        self.clients_cfg = [{'type': 'luna',
                             'api_address': self.cfg.get('api_address'),
                             'db_name': self.cfg.get('db_name'),
                             'max_df_len': self.cfg.get('max_df_len')}]
        self.metrics_objs = {}  # map of case names and metric objects
        self.monitoring_metrics = {}
        self.rps_metrics = {
            'actual_rps_metrics_obj': None,
            'planned_rps_metrics_obj': None,
            'actual_rps_latest': pandas.Series([])
        }
        self.rps_uploader = th.Thread(target=self.upload_planned_rps)

        self._col_map = None
        self._data_session = None
        self._meta = None
        self._test_name = None

    @property
    def meta(self):
        if self._meta is None:
            self._meta = dict(self.get_lp_meta(), **self.cfg.get('meta', {}))
        return self._meta

    @property
    def test_name(self):
        if self._test_name is None:
            self._test_name = self.cfg.get('test_name') or self.core.info.get_value(['uploader', 'job_name'])
        return self._test_name

    def configure(self):
        pass

    def start_test(self):
        try:
            self.reader = self.core.job.generator_plugin.get_reader(parser=string_to_df_microsec)
        except TypeError:
            logger.error('Generator plugin does not support NeUploader')
            self.is_test_finished = lambda: -1
            self.reader = []

    @thread_safe_property
    def col_map(self):
        return {
            'interval_real': self.data_session.new_true_metric,
            'connect_time': self.data_session.new_true_metric,
            'send_time': self.data_session.new_true_metric,
            'latency': self.data_session.new_true_metric,
            'receive_time': self.data_session.new_true_metric,
            'interval_event': self.data_session.new_true_metric,
            'net_code': self.data_session.new_event_metric,
            'proto_code': self.data_session.new_event_metric
        }

    @thread_safe_property
    def data_session(self):
        """
        :rtype: DataSession
        """
        if self._data_session is None:
            config_filenames = {'validated_conf.yaml', 'configinitial.yaml'}
            self._data_session = DataSession({'clients': self.clients_cfg},
                                             tankapi_info=self.tankapi_info(),
                                             config_filenames=config_filenames,
                                             artifacts_dir=self.core.artifacts_dir,
                                             test_start=self.core.info.get_value(['generator', 'test_start'], 0) * 10**6)
            self.add_cleanup(self._cleanup)
            self._data_session.update_job(dict({'name': self.test_name,
                                                '__type': 'tank'},
                                               **self.meta))
            job_no = self._data_session.clients[0].job_number
            if job_no:
                self.publish('job_no', int(job_no))
                self.publish('web_link', urljoin(self.LUNA_LINK, job_no))
        return self._data_session

    def tankapi_info(self):
        meta = self.cfg.get('meta', {})
        return {
            'host': meta.get('tankapi_host'),
            'port': meta.get('tankapi_port'),
            'local_id': self.core.test_id
        }

    def _cleanup(self):
        self.upload_actual_rps(data=pandas.DataFrame([]), last_piece=True)
        uploader_metainfo = self.get_lp_meta()
        autostop_info = self.get_autostop_info()
        regressions = self.get_regressions_names(uploader_metainfo)
        lp_link = self.core.info.get_value(['uploader', 'web_link'])

        meta = self.meta
        meta.update(autostop_info)
        meta['regression'] = regressions
        meta['lunapark_link'] = lp_link

        self.data_session.update_job(meta)
        self.data_session.close(test_end=self.core.info.get_value(['generator', 'test_end'], 0) * 10**6)

    def is_test_finished(self):   # noqa: PLE0202
        df = next(self.reader)
        if df is not None:
            self.upload(df)
        return -1

    def monitoring_data(self, data_list):
        self.upload_monitoring(data_list)

    def post_process(self, retcode):
        try:
            self.rps_uploader.start()
            for chunk in self.reader:
                if chunk is not None:
                    self.upload(chunk)
            self.upload_actual_rps(data=pandas.DataFrame([]), last_piece=True)
            if self.rps_uploader.is_alive():
                self.rps_uploader.join()
        except KeyboardInterrupt:
            logger.warning('Caught KeyboardInterrupt on Neuploader')
            self._cleanup()
        return retcode

    @property
    def is_telegraf(self):
        return True

    def get_metric_obj(self, col, case):
        """
        Generator of metric objects:
        Checks existent metrics and creates new metric if it does not exist.
        :param col:  str with column name
        :param case: str with case name
        :return: metric object
        """

        case_metrics = self.metrics_objs.get(case)
        if case_metrics is None:
            for col, constructor in self.col_map.items():
                self.metrics_objs.setdefault(case, {})[col] = constructor(
                    dict(self.meta,
                         name=col,
                         source='tank',
                         importance='high' if col in self.importance_high else ''),
                    raw=False, aggregate=True,
                    parent=self.get_metric_obj(col, self.OVERALL) if case != self.OVERALL else None,
                    case=case if case != self.OVERALL else None
                )
        return self.metrics_objs[case][col]

    def upload(self, df):
        self.upload_actual_rps(df)

        df_cases_set = set()
        for row in df.itertuples():
            if row.tag and isinstance(row.tag, str):
                df_cases_set.add(row.tag)
                if '|' in row.tag:
                    for tag in row.tag.split('|'):
                        df_cases_set.add(tag)

        for column in self.col_map:
            overall_metric_obj = self.get_metric_obj(column, self.OVERALL)
            df['value'] = df[column]
            result_df = self.filter_df_by_case(df, self.OVERALL)
            overall_metric_obj.put(result_df)

            for case_name in df_cases_set:
                case_metric_obj = self.get_metric_obj(column, case_name)
                df['value'] = df[column]
                result_df = self.filter_df_by_case(df, case_name)
                case_metric_obj.put(result_df)

    def upload_monitoring(self, data):
        for metric_name, df in self.monitoring_data_to_dfs(data).items():
            if metric_name not in self.monitoring_metrics:
                panel, metric = metric_name.split(':', 1)
                try:
                    group, name = metric.split('_', 1)
                except ValueError:
                    name = metric
                    group = '_OTHER_'
                self.monitoring_metrics[metric_name] =\
                    self.data_session.new_true_metric(
                        meta=dict(self.meta,
                                  name=name,
                                  group=group,
                                  host=panel,
                                  type='monitoring'))
            self.monitoring_metrics[metric_name].put(df)

    def upload_planned_rps(self):
        """ Uploads planned rps as a raw metric """
        df = self.parse_stpd()

        if not df.empty:
            self.rps_metrics['planned_rps_metrics_obj'] = self.data_session.new_true_metric(
                meta=dict(self.meta, name=self.PLANNED_RPS_METRICS_NAME, source='tank'),
                raw=True, aggregate=False, parent=None, case=None)
            self.rps_metrics['planned_rps_metrics_obj'].put(df)

    def upload_actual_rps(self, data, last_piece=False):
        """ Upload actual rps metric """
        if self.rps_metrics['actual_rps_metrics_obj'] is None:
            self.rps_metrics['actual_rps_metrics_obj'] = self.data_session.new_true_metric(
                meta=dict(self.meta, name=self.ACTUAL_RPS_METRICS_NAME),
                raw=True, aggregate=False, parent=None, case=None
            )
        df = self.count_actual_rps(data, last_piece)
        if not df.empty:
            self.rps_metrics['actual_rps_metrics_obj'].put(df)

    def parse_stpd(self):
        """  Reads rps plan from stpd file """
        stpd_file = self.core.info.get_value(['stepper', 'stpd_file'])
        if not stpd_file:
            logger.info('No stpd found, no planned_rps metrics')
            return pandas.DataFrame()

        rows_list = []
        test_start = int(self.core.info.get_value(['generator', 'test_start'], 0) * 10 ** 3)
        pattern = r'^\d+ (\d+)\s*.*$'
        regex = re.compile(pattern)
        try:
            with open(stpd_file) as stpd:
                for line in stpd:
                    if regex.match(line):
                        timestamp = int((int(line.split(' ')[1]) + test_start) / 1e3)  # seconds
                        rows_list.append(timestamp)
        except Exception:
            logger.warning('Failed to parse stpd file')
            logger.debug('', exc_info=True)
            return pandas.DataFrame()

        return self.rps_series_to_df(pandas.Series(rows_list))

    def count_actual_rps(self, data, last_piece):
        """ Counts actual rps on base of input chunk. Uses buffer for latest timestamp in df. """
        if not last_piece and not data.empty:
            concat_ts = pandas.concat([(data.ts / 1e6).astype(int), self.rps_metrics['actual_rps_latest']])
            self.rps_metrics['actual_rps_latest'] = concat_ts.loc[lambda s: s == concat_ts.max()]
            series_to_send = concat_ts.loc[lambda s: s < concat_ts.max()]
            df = self.rps_series_to_df(series_to_send) if series_to_send.any else pandas.DataFrame([])
        else:
            df = self.rps_series_to_df(self.rps_metrics['actual_rps_latest'])
            self.rps_metrics['actual_rps_latest'] = pandas.Series()
        return df

    @staticmethod
    def monitoring_data_to_dfs(data):
        panels = {}
        for chunk in data:
            for panel_name, content in chunk['data'].items():
                if panel_name in panels:
                    for metric_name, value in content['metrics'].items():
                        if metric_name in panels[panel_name]:
                            panels[panel_name][metric_name]['value'].append(value)
                            panels[panel_name][metric_name]['ts'].append(chunk['timestamp'])
                        else:
                            panels[panel_name][metric_name] = {'value': [value], 'ts': [chunk['timestamp']]}
                else:
                    panels[panel_name] = {name: {'value': [value], 'ts': [chunk['timestamp']]} for name, value in content['metrics'].items()}
        return {'{}:{}'.format(panelk, name): pandas.DataFrame({'ts': [ts * 1000000 for ts in values['ts']], 'value': values['value']})
                for panelk, panelv in panels.items() for name, values in panelv.items()}

    @staticmethod
    def rps_series_to_df(series):
        df = series.value_counts().to_frame(name='value')
        df_to_send = df.rename_axis('ts')
        df_to_send.reset_index(inplace=True)
        df_to_send.loc[:, 'ts'] = (df_to_send['ts'] * 1e6).astype(int)
        return df_to_send

    @staticmethod
    def filter_df_by_case(df, case):
        """
        Filter dataframe by case name. If case is '__overall__', return all rows.
        :param df: DataFrame
        :param case: str with case name
        :return: DataFrame with columns 'ts' and 'value'
        """
        case = case.strip()
        return df[['ts', 'value']] if case == Plugin.OVERALL else df[df.tag.str.strip() == case][['ts', 'value']]

    def get_lp_meta(self):
        uploader_meta = self.core.info.get_value(['uploader'])
        if not uploader_meta:
            logger.info('No uploader metainfo found')
            return {}
        else:
            meta_tags_names = ['component', 'description', 'name', 'person', 'task', 'version', 'lunapark_jobno']
            meta_tags = {key: uploader_meta.get(key) for key in meta_tags_names if key in uploader_meta}
            meta_tags.update({k: v if v is not None else '' for k, v in uploader_meta.get('meta', {}).items()})
            return meta_tags

    @staticmethod
    def get_regressions_names(uploader_metainfo):
        task, component_name = uploader_metainfo.get('task'), uploader_metainfo.get('component')
        if not task or not component_name:
            return []
        project_name = task.split('-')[0]
        lp_api_url = 'https://lunapark.yandex-team.ru/api/regress/{}/componentlist.json'.format(project_name)
        try:
            componentlist =\
                requests.get(lp_api_url).json()
        except (ValueError, ConnectionError):
            logger.info("Failed to fetch data from {}".format(lp_api_url), exc_info=True)
            return []
        for component in componentlist:
            try:
                if component['name'] == component_name:
                    services = component['services']
                    if len(services) == 0:
                        services = ['__OTHER__']
                    return ['{}_{}'.format(project_name, s).replace(' ', '_') for s in services]
            except KeyError:
                pass
        else:
            return []

    def get_autostop_info(self):
        autostop_info = self.core.info.get_value(['autostop'])
        if autostop_info:
            autostop_rps = autostop_info.get('rps', 0)
            autostop_reason = autostop_info.get('reason', '')
            self.log.warning('Autostop: %s %s', autostop_rps, autostop_reason)
            return {'autostop_rps': autostop_rps, 'autostop_reason': autostop_reason}
        else:
            return {}



from setuptools import setup, find_packages

setup(
    name='netort',
    version='0.8.2',
    description='common library for yandex-load org',
    author='Timur Torubarov (load testing)',
    author_email='netort@yandex-team.ru',
    maintainer='Yandex load team',
    maintainer_email='load@yandex-team.ru',
    url='http://github.com/yandex-load/netort',
    packages=find_packages(exclude=["tests", "tmp", "docs", "data"]),
    install_requires=[
        'pyserial', 'requests', 'retrying', 'cerberus', 'six>=1.12.0', 'pandas<=1.5.3',
        'pathlib', 'typing', 'numpy'
    ],
    setup_requires=[
        # 'pytest-runner', 'flake8',
    ],
    tests_require=[
        'pytest', 'pytest-runner', 'mock'
    ],
    license='MPLv2',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Environment :: Web Environment',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: GNU Lesser General Public License v2 or later (LGPLv2+)',
        'Operating System :: POSIX',
        'Topic :: Software Development :: Quality Assurance',
        'Topic :: Software Development :: Testing',
        'Topic :: Software Development :: Testing :: Traffic Generation',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
    ],
    entry_points={
        'console_scripts': [
            'phout_upload = netort.cli:main',
        ],
    },
    package_data={
    },
    use_2to3=False, )



import datetime
import logging
import subprocess
import time
import os
import re
import shutil
from threading import Event

import yaml

from yandextank.contrib.netort.netort.resource import manager as resource_manager
from yandextank.contrib.netort.netort.resource import HttpOpener

from .reader import PandoraStatsReader
from ..Console import Plugin as ConsolePlugin
from ..Console import screen as ConsoleScreen
from ..Phantom import PhantomReader, string_to_df
from ...common.interfaces import AbstractInfoWidget, GeneratorPlugin
from ...common.util import tail_lines, FileMultiReader

logger = logging.getLogger(__name__)


class Plugin(GeneratorPlugin):
    """    Pandora load generator plugin    """

    OPTION_CONFIG = "config"
    SECTION = "pandora"
    DEFAULT_REPORT_FILE = "phout.log"
    DEFAULT_EXPVAR_PORT = 1234

    def __init__(self, core, cfg, name):
        super(Plugin, self).__init__(core, cfg, name)
        self.output_finished = Event()
        self.enum_ammo = False
        self.pandora_cmd = None
        self.pandora_config_file = None
        self.config_contents = None
        self.custom_config = False
        self.expvar = self.get_option('expvar')
        self.expvar_enabled = self.expvar
        self.expvar_port = self.DEFAULT_EXPVAR_PORT
        self.report_files = None
        self.__address = None
        self.__schedule = None
        self.ammofile = None
        self.process_stderr_file = None
        self.resources = []

    @staticmethod
    def get_key():
        return __file__

    def get_available_options(self):
        opts = [
            "pandora_cmd", "buffered_seconds",
            "config_content", "config_file"
        ]
        return opts

    def configure(self):
        self.report_files = [self.get_option("report_file")]
        self.buffered_seconds = self.get_option("buffered_seconds")
        self.affinity = self.get_option("affinity", "")
        self.resources = self.get_option("resources")

        # if we use custom pandora binary, we can download it and make it executable
        self.pandora_cmd = self.get_resource(self.get_option("pandora_cmd"), "./pandora", permissions=0o755)

        # download all resources from self.get_options("resources")
        if len(self.resources) > 0:
            for resource in self.resources:
                self.get_resource(resource["src"], resource["dst"])

        # get config_contents and patch it: expand resources via resource manager
        # config_content option has more priority over config_file
        if self.get_option("config_content"):
            logger.info('Found config_content option configuration')
            self.config_contents = self.__patch_raw_config_and_dump(self.get_option("config_content"))
        elif self.get_option("config_file"):
            logger.info('Found config_file option configuration')
            with open(self.get_option("config_file"), 'rb') as config:
                external_file_config_contents = yaml.load(config.read(), Loader=yaml.FullLoader)
            self.config_contents = self.__patch_raw_config_and_dump(external_file_config_contents)
        else:
            raise RuntimeError("Neither pandora.config_content, nor pandora.config_file specified")
        logger.debug('Config after parsing for patching: %s', self.config_contents)

        # find report filename and add to artifacts
        self.report_files = [pool['result']['destination'] for pool in self.config_contents['pools']]
        for f in self.report_files:
            with open(f, 'w'):
                pass
            self.core.add_artifact_file(f)

    def __patch_raw_config_and_dump(self, cfg_dict):
        if not cfg_dict:
            raise RuntimeError('Empty pandora config')
        # patch
        config_content = self.patch_config(cfg_dict)
        # dump
        self.pandora_config_file = self.core.mkstemp(".yaml", "pandora_config_")
        self.core.add_artifact_file(self.pandora_config_file)
        with open(self.pandora_config_file, 'w') as config_file:
            yaml.dump(config_content, config_file)
        return config_content

    def patch_config(self, config):
        """
        download remote resources, replace links with local filenames
        add result file section
        :param dict config: pandora config
        """
        # get expvar parameters
        if config.get("monitoring"):
            if config["monitoring"].get("expvar"):
                self.expvar_enabled = config["monitoring"]["expvar"].get("enabled")
                if config["monitoring"]["expvar"].get("port"):
                    self.expvar_port = config["monitoring"]["expvar"].get("port")
        # or set if expvar not exists
        elif not self.expvar:
            config["monitoring"] = {
                "expvar": {
                    "enabled": True,
                }
            }
            self.expvar_enabled = True

        # FIXME this is broken for custom ammo providers due to interface incompatibility
        # FIXME refactor pandora plx
        for n, pool in enumerate(config['pools']):
            if pool.get('ammo', {}).get('file', ''):
                self.ammofile = pool['ammo']['file']
                opener = resource_manager.get_opener(self.ammofile)
                if isinstance(opener, HttpOpener):
                    pool['ammo']['file'] = opener.download_file(True, try_ungzip=True)
                else:
                    pool['ammo']['file'] = opener.get_filename

            if not pool.get('result') or 'phout' not in pool.get('result', {}).get('type', ''):
                logger.warning('Seems like pandora result file not specified... adding defaults')
                pool['result'] = dict(
                    destination=f"{n}_{self.DEFAULT_REPORT_FILE}",
                    type='phout',
                )

            if pool.get('gun').get('answlog'):
                answ = pool.get('gun').get('answlog')
                if answ.get('enabled') is True:
                    if not answ.get('path'):
                        answ_path = self.core.mkstemp(".log", "answ_")
                        logger.warning('Seems like pandora answer log file not specified... adding default path: %s', answ_path)
                        self.core.add_artifact_file(answ_path)
                        config['pools'][n]['gun']['answlog']['path'] = answ_path
                    else:
                        answ_path = answ.get('path')
                        self.core.add_artifact_file(answ_path)

        return config

    @property
    def address(self):
        if not self.__address:
            for pool in self.config_contents['pools']:
                if pool.get('gun', {}).get('target'):
                    self.__address = pool.get('gun', {}).get('target')
                    break
            else:
                self.__address = 'unknown'
        return self.__address

    @property
    def schedule(self):
        if not self.__schedule:
            for pool in self.config_contents['pools']:
                if pool.get('rps'):
                    self.__schedule = pool.get('rps')
                    break
            else:
                self.__schedule = 'unknown'
        return self.__schedule

    def get_info(self):
        return self.Info(
            address=self.address,
            ammo_file=self.ammofile,
            duration=0,
            instances=0,
            loop_count=0,
            port=self.address.split(':')[-1],
            rps_schedule=self.schedule
        )

    def get_reader(self, parser=string_to_df):
        if self.reader is None:
            self.reader = [FileMultiReader(f, self.output_finished) for f in self.report_files]
        return [PhantomReader(reader.get_file(), parser=parser) for reader in self.reader]

    def get_stats_reader(self):
        if self.stats_reader is None:
            self.stats_reader = PandoraStatsReader(self.expvar_enabled, self.expvar_port)
        return self.stats_reader

    def get_resource(self, resource, dst, permissions=0o644):
        opener = resource_manager.get_opener(resource)
        if isinstance(opener, HttpOpener):
            tmp_path = opener.download_file(True, try_ungzip=True)
            shutil.copy(tmp_path, dst)
            logger.info('Successfully moved resource %s', dst)
        else:
            dst = opener.get_filename
        try:
            os.chmod(dst, permissions)
        except OSError:
            logger.warning('Cannot change permissions to %s', dst)
        return dst

    def prepare_test(self):
        try:
            console = self.core.get_plugin_of_type(ConsolePlugin)
        except KeyError as ex:
            logger.debug(ex)
            console = None

        if console:
            widget = PandoraInfoWidget(self)
            console.add_info_widget(widget)
            self.core.job.aggregator.add_result_listener(widget)

    def start_test(self):
        args = [self.pandora_cmd] +\
            (['-expvar'] if self.expvar else []) +\
            [self.pandora_config_file]
        if self.affinity:
            self.core.__setup_affinity(self.affinity, args=args)
        logger.info("Starting: %s", args)
        self.start_time = time.time()
        self.process_stderr_file = self.core.mkstemp(".log", "pandora_")
        self.core.add_artifact_file(self.process_stderr_file)
        self.process_stderr = open(self.process_stderr_file, 'w')
        try:
            self.process = subprocess.Popen(
                args,
                stderr=self.process_stderr,
                stdout=self.process_stderr,
                close_fds=True)
        except OSError:
            logger.debug(
                "Unable to start Pandora binary. Args: %s", args, exc_info=True)
            raise RuntimeError(
                "Unable to start Pandora binary and/or file does not exist: %s" % args)

    def is_test_finished(self):
        retcode = self.process.poll()
        if retcode is not None and retcode == 0:
            logger.info("Pandora subprocess done its work successfully and finished w/ retcode 0")
            self.output_finished.set()
            return retcode
        elif retcode is not None and retcode != 0:
            lines_amount = 20
            logger.info("Pandora finished with non-zero retcode. Last %s logs of Pandora log:", lines_amount)
            self.output_finished.set()
            last_log_contents = tail_lines(self.process_stderr_file, lines_amount)
            for logline in last_log_contents:
                logger.info(logline.strip('\n'))
            with open(self.process_stderr_file) as stderr:
                for line in stderr:
                    if self.check_log_line_contains_error(line):
                        self.errors.append(line.strip())
            return abs(retcode)
        else:
            return -1

    @staticmethod
    def check_log_line_contains_error(line):
        return re.search('^panic:|ERROR|FATAL', line) is not None

    def end_test(self, retcode):
        if self.process and self.process.poll() is None:
            logger.warning(
                "Terminating worker process with PID %s", self.process.pid)
            self.process.terminate()
            if self.process_stderr:
                self.process_stderr.close()
        else:
            logger.debug("Seems subprocess finished OK")
        self.output_finished.set()
        return retcode


class PandoraInfoWidget(AbstractInfoWidget):
    ''' Right panel widget '''

    def __init__(self, pandora):
        AbstractInfoWidget.__init__(self)
        self.krutilka = ConsoleScreen.krutilka()
        self.owner = pandora
        self.reqps = 0
        self.active = 0

    def get_index(self):
        return 0

    def on_aggregated_data(self, data, stats):
        self.reqps = stats["metrics"]["reqps"]
        self.active = stats["metrics"]["instances"]

    def render(self, screen):
        text = " Pandora Test %s" % next(self.krutilka)
        space = screen.right_panel_width - len(text) - 1
        left_spaces = space // 2
        right_spaces = space // 2

        dur_seconds = int(time.time()) - int(self.owner.start_time)
        duration = str(datetime.timedelta(seconds=dur_seconds))

        template = screen.markup.BG_BROWN + '~' * left_spaces + \
            text + ' ' + '~' * right_spaces + screen.markup.RESET + "\n"
        template += "Command Line: %s\n"
        template += "    Duration: %s\n"
        template += "  Requests/s: %s\n"
        template += " Active reqs: %s\n"
        template += "      Target: %s\n"
        template += "    Schedule: \n%s\n"
        data = (
            self.owner.pandora_cmd,
            duration,
            self.reqps,
            self.active,
            self.owner.address,
            yaml.dump(self.owner.schedule)
        )

        return template % data



import errno
import collections
import logging
import os.path
import subprocess
import time

from ...common.interfaces import AbstractPlugin, GeneratorPlugin, AggregateResultListener, AbstractInfoWidget, \
    StatsReader
from ...common.util import FileScanner
from ..Console import Plugin as ConsolePlugin
from ..Phantom import PhantomReader


_INFO = collections.namedtuple(
    "Info",
    "address, port, instances, ammo_count, loop_count, duration, steps, stat_log, rps_schedule, ammo_file"
)
_LOGGER = logging.getLogger(__name__)

_PROCESS_KILL_TIMEOUT = 10  # Kill running process after specified number of seconds
_OUTPUT_WAIT_TIMEOUT = 10  # Output files should be found after specified number of seconds


class Plugin(GeneratorPlugin):
    """Simple executor of shooting process with phantom compatible output"""

    SECTION = 'shootexec'

    def __init__(self, core, cfg, name):
        AbstractPlugin.__init__(self, core, cfg, name)
        self.stats_reader = None
        self.reader = None
        self.__process = None
        self.__stderr_file = None
        self.__processed_ammo_count = 0
        self.start_time = 0
        self.opened_file = None

    @staticmethod
    def get_key():
        return __file__

    def get_available_options(self):
        return ["cmd", "output_path", "stats_path"]

    def configure(self):
        self.__cmd = self.get_option("cmd")

        self.__output_path = self.get_option("output_path")
        self.core.add_artifact_file(self.__output_path)

        self.__stats_path = self.get_option("stats_path")
        if self.__stats_path:
            self.core.add_artifact_file(self.__stats_path)

    def get_reader(self):
        if self.reader is None:
            # Touch output_path to clear it
            open(self.__output_path, "w").close()
            self.opened_file = open(self.__output_path, 'r')
            self.add_cleanup(lambda: self.opened_file.close())
            self.reader = PhantomReader(self.opened_file)
        return self.reader

    def get_stats_reader(self):
        if self.stats_reader is None:
            self.stats_reader = _FileStatsReader(self.__stats_path) if self.__stats_path else _DummyStatsReader()
        return self.stats_reader

    def prepare_test(self):
        stderr_path = self.core.mkstemp(".log", "shootexec_stdout_stderr_")
        self.__stderr_file = open(stderr_path, 'w')

        _LOGGER.debug("Linking sample reader to aggregator. Reading samples from %s", self.__output_path)

        self.start_time = time.time()
        self.core.job.aggregator.add_result_listener(self)

        try:
            console = self.core.get_plugin_of_type(ConsolePlugin)
        except KeyError as ex:
            _LOGGER.debug(ex)
            console = None

        if console:
            widget = _InfoWidget(self)
            console.add_info_widget(widget)
            self.core.job.aggregator.add_result_listener(widget)

    def start_test(self):
        _LOGGER.info("Starting shooting process: '%s'", self.__cmd)
        self.__process = subprocess.Popen(
            self.__cmd,
            shell=True,
            stderr=self.__stderr_file,
            stdout=self.__stderr_file,
            close_fds=True
        )

        # Ensure that all expected output files are ready to use
        _LOGGER.info("Waiting until output files are ready")
        waitfor = time.time() + _OUTPUT_WAIT_TIMEOUT
        while time.time() < waitfor:
            output_path_is_ready = os.path.isfile(self.__output_path)
            stats_path_is_ready = (not self.__stats_path or os.path.isfile(self.__stats_path))
            if output_path_is_ready and stats_path_is_ready:
                break
            time.sleep(0.1)
        else:
            raise Exception("Failed to wait until output resources are ready: output={}, stats={}".format(
                output_path_is_ready,
                stats_path_is_ready
            ))
        _LOGGER.info("Shooting proces is ready to use")

    def is_test_finished(self):
        retcode = self.__process.poll()
        if retcode is not None:
            _LOGGER.info("Shooting process done its work with exit code: %s", retcode)
            return abs(retcode)
        else:
            return -1

    def end_test(self, retcode):
        if self.__process and self.__process.poll() is None:
            _LOGGER.warn("Terminating shooting process with PID %s", self.__process.pid)
            self.__terminate()
        else:
            _LOGGER.debug("Seems shooting process finished OK")
        return retcode

    def post_process(self, retcode):
        return retcode

    def on_aggregated_data(self, data, stats):
        self.__processed_ammo_count += data["overall"]["interval_real"]["len"]
        _LOGGER.debug("Processed ammo count: %s", self.__processed_ammo_count)

    def get_info(self):
        """ returns info object """
        return _INFO(
            "",
            "",
            "0",
            "0",
            "0",
            time.time() - self.start_time,
            None,
            "",
            "",
            ""
        )

    def __terminate(self):
        """Gracefull termination of running process"""

        if self.__stderr_file:
            self.__stderr_file.close()

        if not self.__process:
            return

        waitfor = time.time() + _PROCESS_KILL_TIMEOUT
        while time.time() < waitfor:
            try:
                self.__process.terminate()
            except EnvironmentError as e:
                if e.errno != errno.ESRCH:
                    _LOGGER.warning("Failed to terminate process '{}': {}".format(self.__cmd, e))
                return
            time.sleep(0.1)

        try:
            self.__process.kill()
        except EnvironmentError as e:
            if e.errno != errno.ESRCH:
                _LOGGER.warning("Failed to kill process '{}': {}".format(self.__cmd, e))
            return


class _InfoWidget(AbstractInfoWidget, AggregateResultListener):
    """
    Widget with information about current run state
    """

    def get_index(self):
        return 2

    def __init__(self, sender):
        AbstractInfoWidget.__init__(self)
        self.owner = sender

    def render(self, screen):
        return ""

    def on_aggregated_data(self, data, stats):
        pass


class _FileStatsReader(FileScanner, StatsReader):
    """
    Read shooting stats line by line

    Line format is 'timestamp\trps\tinstances'
    """

    def __init__(self, *args, **kwargs):
        super(_FileStatsReader, self).__init__(*args, **kwargs)
        self.__last_ts = 0

    def _read_data(self, lines):
        """
        Parse lines and return stats
        """

        results = []
        for line in lines:
            timestamp, rps, instances = line.split("\t")
            curr_ts = int(float(timestamp))  # We allow floats here, but tank expects only seconds
            if self.__last_ts < curr_ts:
                self.__last_ts = curr_ts
                results.append(self.stats_item(self.__last_ts, float(rps), float(instances)))
        return results


class _DummyStatsReader(StatsReader):
    """
    Dummy stats reader for shooters without stats file
    """

    def __init__(self):
        self.__closed = False
        self.__last_ts = 0

    def __iter__(self):
        while not self.__closed:
            cur_ts = int(time.time())
            if cur_ts > self.__last_ts:
                yield [self.stats_item(cur_ts, 0, 0)]
                self.__last_ts = cur_ts
            else:
                yield []

    def close(self):
        self.__closed = True



from .solomonsensor import SolomonSensor
from logging import getLogger
from multiprocessing import Queue
from re import match, findall
from threading import Thread, Event
from time import sleep, time


logger = getLogger(__name__)


def convert_name(name):
    NAME_MAP = {r'conv\(.+\)': r'(?<=conv\().+?(?=,)'}
    for pattern, mask in NAME_MAP.items():
        if match(pattern, name):
            name = findall(mask, name)[0]
            break
    return 'custom:{}'.format(name)[:100]


def monitoring_data(pname, metrics, comment=""):
    try:
        return {
            "timestamp": tuple(metrics)[0],
            "data": {
                pname: {
                    "comment": comment,
                    "metrics": {convert_name(name): value for name, value in tuple(metrics.values())[0].items()}}}}
    except (IndexError, ValueError) as ve:
        logger.warning("Can't format. Wrong data {}. {}".format(metrics, ve), exc_info=True)


class SolomonReceiver(object):

    """
    The SolomonReceiver class collects metrics for all sensors and aggregates them together.
    """

    def __init__(self, api_host, config, timeout, token):
        self.api_host = api_host
        self.config = config
        self.token = token
        self.stop = Event()
        self.timeout = timeout
        self.sensors = list()
        self.panels = list()
        self._data_buffer = list()

    def get_buffered_data(self):
        data, self._data_buffer = self._data_buffer, []
        return data

    def prepare(self):
        for name, data in self.config.items():
            panel = Panel(name, self.timeout)
            senset = set()
            for dto in data['sensors']:
                sensor = SolomonSensor(self.api_host, data['project'], dto, panel.queue, self.token, data['priority_labels'])
                senset.update(sensor.get_sensors())
                self.sensors.append(Thread(target=self.run_sensor, args=(sensor,)))
            panel.set_sensors(senset)
            self.panels.append(Thread(target=self.run_panel, args=(panel,)))

    def run(self):
        for thread in self.panels:
            if not thread.is_alive():
                thread.start()
        for thread in self.sensors:
            if not thread.is_alive():
                thread.start()

    def stop_now(self):
        logger.info("Waiting for finish the Solomon plugin after {} second".format(self.timeout))
        sleep(self.timeout + 1)
        self.stop.set()
        for thread in self.sensors:
            if thread.is_alive():
                thread.join()
        for thread in self.panels:
            if thread.is_alive():
                thread.join()

    def run_sensor(self, sensor):
        while not self.stop.is_set() and not sensor.queue._closed:
            sleep(self.timeout)
            sensor.get_metrics()

    def run_panel(self, panel):
        try:
            while not self.stop.is_set():
                panel.process_queue()
                for metric in panel.ready_metrics():
                    self._data_buffer.append(monitoring_data(panel.name, metric))
                sleep(5)
        except Exception as ex:
            logger.error("Error {} for Solomon panel {}. {}".format(type(ex), panel.name, ex), exc_info=True)
        finally:
            logger.info("Solomon panel {} will be closed after {} second".format(panel.name, self.timeout))
            sleep(self.timeout)
            panel.process_queue()
            for ts, metrics in panel.buffer.items():
                self._data_buffer.append(monitoring_data(panel.name, {ts: metrics}))


class Panel(object):

    """
    The class Panel is used for initializing and managing structural levels for collecting monitoring data passed through the configuration file.
    """

    def __init__(self, name, timeout):
        self.name = name
        self.queue = Queue()
        self.timeout = timeout
        self.sensors = set()
        self.buffer = dict()

    def set_sensors(self, sensors):
        self.sensors = sensors

    def process_queue(self):
        try:
            while not self.queue.empty():
                self.compose_metrics(self.queue.get())
        except (IOError, OSError) as error:
            logger.warning("Panel {} get metrics from queue error. {}".format(self.name, error), exc_info=True)

    def compose_metrics(self, metric):
        if isinstance(metric, dict) and 'timestamp' in metric:
            if metric['timestamp'] in self.buffer:
                self.buffer[metric['timestamp']].update({metric['sensor']: metric['value']})
            else:
                self.buffer[metric['timestamp']] = {metric['sensor']: metric['value']}

    def ready_metrics(self):
        now = time()
        outbuffer = list()
        complite_ts = list()
        for ts, metrics in self.buffer.items():
            if self.sensors.issubset(set(metrics.keys())) or (now - ts) > self.timeout:
                outbuffer.append({ts: metrics})
                complite_ts.append(ts)
        for ts in complite_ts:
            self.buffer.pop(ts)
        return outbuffer



import logging
import time
from threading import Event
import subprocess

import yaml

from .reader import BfgStatsReader
from .widgets import BfgInfoWidget
from ..Phantom import PhantomReader, string_to_df
from ..Console import Plugin as ConsolePlugin
from ...common.interfaces import GeneratorPlugin
from ...common.util import FileMultiReader, tail_lines
from ...stepper import StepperWrapper


class Plugin(GeneratorPlugin):
    """ internal Big Fucking Gun plugin """
    SECTION = 'bfg'

    def __init__(self, core, cfg, name):
        super(Plugin, self).__init__(core, cfg, name)
        self.close_event = Event()
        self.bfg_cmd = self.get_option("bfg_cmd")
        self.bfg_config_file = self.core.mkstemp(".yaml", "bfg_config_")
        self.log = logging.getLogger(__name__)
        self.stepper_wrapper = StepperWrapper(core, cfg)
        self.log.info("Initialized BFG")
        self.report_filename = "phout.log"

    @staticmethod
    def get_key():
        return __file__

    def get_available_options(self):
        return [
            "gun_type", "instances", "cached_stpd", "pip"
        ]

    def configure(self):
        self.__dump_config()
        self.core.add_artifact_file(self.report_filename)
        with open(self.report_filename, "a"):
            self.log.debug("Bfg2020 phout file created")

    def __dump_config(self):
        config_content = self.core.config.validated["bfg"]
        cache_dir = self.get_option("cache_dir") or self.core.artifacts_base_dir
        config_content["cache_dir"] = cache_dir
        self.core.add_artifact_file(self.bfg_config_file)
        with open(self.bfg_config_file, 'w') as config_file:
            yaml.dump(config_content, config_file)

    def get_reader(self, parser=string_to_df):
        if self.reader is None:
            self.reader = FileMultiReader(self.report_filename, self.close_event)
        return PhantomReader(self.reader.get_file(), parser=parser)

    def get_stats_reader(self):
        if self.stats_reader is None:
            self.stats_reader = BfgStatsReader("1234")
        return self.stats_reader

    def prepare_test(self):
        try:
            console = self.core.get_plugin_of_type(ConsolePlugin)
        except KeyError as ex:
            self.log.debug(ex)
            console = None

        if console:
            widget = BfgInfoWidget()
            console.add_info_widget(widget)
            self.core.job.aggregator.add_result_listener(widget)

    def start_test(self):
        args = [self.bfg_cmd] +\
            ['--config=' + self.bfg_config_file]
        self.log.info("Starting: %s", args)
        self.start_time = time.time()
        self.process_stderr_file = self.core.mkstemp(".log", "bfg_")
        self.core.add_artifact_file(self.process_stderr_file)
        self.process_stderr = open(self.process_stderr_file, 'w')
        try:
            self.process = subprocess.Popen(
                args,
                stderr=self.process_stderr,
                stdout=self.process_stderr,
                close_fds=True)
        except OSError:
            self.log.debug(
                "Unable to start Bfg binary. Args: %s", args, exc_info=True)
            raise RuntimeError(
                "Unable to start Bfg binary and/or file does not exist: %s" % args)

    def is_test_finished(self):
        retcode = self.process.poll()
        if retcode is not None and retcode == 0:
            self.log.info("BFG subprocess done its work successfully and finished w/ retcode 0")
            self.close_event.set()
            return retcode
        elif retcode is not None and retcode != 0:
            lines_amount = 20
            self.log.info("BFG finished with non-zero retcode. Last %s logs of BFG log:", lines_amount)
            self.close_event.set()
            last_log_contents = tail_lines(self.process_stderr_file, lines_amount)
            for logline in last_log_contents:
                self.log.info(logline.strip('\n'))
            return abs(retcode)
        else:
            return -1

    def end_test(self, retcode):
        if self.process and self.process.poll() is None:
            self.log.warning(
                "Terminating worker process with PID %s", self.process.pid)
            self.process.terminate()
            if self.process_stderr:
                self.process_stderr.close()
        else:
            self.log.debug("Subprocess finished")
        self.close_event.set()
        return retcode



from datetime import datetime
from json import loads
from logging import getLogger
from requests import post
from time import time


logger = getLogger(__name__)


class SolomonSensor(object):
    """
    The SolomonSensor class is intended for obtaining actual discrete values and composing these data into a convenient format for further processing.
    """

    GET_DATA_URI = '/api/v2/projects/{}/sensors/data'

    def __init__(self, api_host, prj, dto, queue, token, priority_labels):
        self.prj = prj
        self.body = self.get_body(dto)
        self.headers = {'Content-Type': "text/plain", 'Accept': "application/json", 'Authorization': token}
        self.data_url = "http://{0}{1}".format(api_host, self.GET_DATA_URI.format(prj))
        self.fromtime = time()
        self.queue = queue
        self.priority_labels = priority_labels
        self.metric_labels = ['app', 'metric', 'name', 'path', 'sensor', 'signal']
        self.ignore_labels = ['cluster', 'project', 'service', 'servant', 'os'] + self.metric_labels + self.priority_labels

    def get_metrics(self):
        now = time()
        fromtime = datetime.utcfromtimestamp(self.fromtime - 5).isoformat()
        totime = datetime.utcfromtimestamp(now).isoformat()
        self.fromtime = now
        self.parse_metrics(self.get_data(fromtime, totime))

    def parse_metrics(self, data):
        if isinstance(data, dict) and 'vector' in data.keys():
            try:
                for vector in data['vector']:
                    labels = vector['timeseries']['labels']
                    values = vector['timeseries']['values']
                    if 'timestamps' in vector['timeseries'].keys():
                        timestamps = vector['timeseries']['timestamps']
                    else:
                        timestamps = list()
                    self.send_metrics(labels, timestamps, values)
            except KeyError as ke:
                logger.warning("Wrong data: {}. {}".format(data, ke), exc_info=True)
        else:
            logger.warning("Wrong data: {}".format(data))

    def send_metrics(self, labels, timestamps, values):
        sensor = self.format_sensor(labels)
        try:
            for order, ts in enumerate(timestamps):
                self.queue.put(
                    {
                        'sensor': sensor,
                        'timestamp': int(ts) / 1000,
                        'value': values[order]
                    }
                )
        except (IOError, OSError) as error:
            logger.warning("Sensor {} send metrics error. {}".format(self.body, error), exc_info=True)

    def get_data(self, fromtime, totime):
        params = {
            'maxPoints': 500,
            'from': "{}Z".format(fromtime),
            'to': "{}Z".format(totime)
        }
        try:
            response = post(self.data_url, params=params, headers=self.headers, data=self.body)
        except ConnectionError as ce:
            logger.error("Error during request. {}".format(ce), exc_info=True)
        else:
            if response.status_code != 200:
                logger.warning('Wrong response code {} for body {}. {}'.format(response.status_code, self.body, response.json()))
                return None
            else:
                return response.json()

    def get_sensors(self):
        sensors = set()
        now = time()
        fromtime = datetime.utcfromtimestamp(now - 1).isoformat()
        totime = datetime.utcfromtimestamp(now).isoformat()
        data = self.get_data(fromtime, totime)
        if isinstance(data, dict) and 'vector' in data.keys():
            for vector in data['vector']:
                sensors.add(self.format_sensor(vector['timeseries']['labels']))
        return sensors

    def format_sensor(self, data):
        if isinstance(data, dict):
            for label in self.metric_labels:
                if label in data.keys():
                    sensor = data[label].replace("/", "-").replace(".", "-")
                    break
            else:
                logger.warning("Wrong labels {}".format(data))
                sensor = "Summary"
            for label in self.priority_labels:
                if label in data.keys():
                    sensor += "-{}".format(data[label].strip(r"\-.*_/").replace("/", "-").replace(".", "-"))
            for key, value in data.items():
                if key in self.ignore_labels:
                    continue
                else:
                    sensor += "_{}".format(str(value).strip(r"\-.*_/").replace("/", "-").replace(".", "-"))
            return sensor
        else:
            return None

    def get_body(self, data):
        if isinstance(data, dict):
            body = self.make_body(data)
        elif isinstance(data, str):
            try:
                body = self.make_body(loads(data))
            except ValueError:
                body = data.strip('\'\\')
        else:
            body = ""
        return body

    def make_body(self, dto):
        bodylist = list()
        if isinstance(dto, dict):
            for key, value in dto.items():
                bodylist.append("{}=\"{}\"".format(key, value))
        return "{{{bl}}}".format(bl=", ".join(bodylist))
